"""
Base classes for invoking a Python script from within
the current Python process.
"""

import collections
import contextlib
import io
import sys

SavedIO = collections.namedtuple('SavedIO', 'stdout stderr')


class TextIO(io.StringIO):
    def write(self, data):
        if not isinstance(data, str):
            data = str(data, getattr(self, '_encoding', 'UTF-8'), 'replace')
        io.StringIO.write(self, data)


@contextlib.contextmanager
def capture_stdio():
    sys_stdout, sys.stdout = sys.stdout, TextIO()
    sys_stderr, sys.stderr = sys.stderr, TextIO()
    try:
        yield SavedIO(sys.stdout, sys.stderr)
    finally:
        sys.stdout = sys_stdout
        sys.stderr = sys_stderr


@contextlib.contextmanager
def replace_sysargv(params):
    sys_argv, sys.argv = sys.argv, params
    try:
        yield
    finally:
        sys.argv = sys_argv


class Result:
    pass


@contextlib.contextmanager
def capture_system_exit():
    res = Result()
    try:
        yield res
        res.code = 0
    except SystemExit as e:
        if isinstance(e.code, int):
            res.code = e.code
        else:
            res.code = 1
    except BaseException:
        res.code = 1
        raise


class ProcessResult:
    pass


@contextlib.contextmanager
def in_process_context(params):
    res = ProcessResult()
    try:
        with capture_stdio() as stdio:
            with replace_sysargv(params):
                with capture_system_exit() as proc_res:
                    yield res
    finally:
        res.stdio = stdio
        res.returncode = proc_res.code



"""
This project implements several repo implementations, each of which provides
an interface to source code repository functionality.
"""

from __future__ import annotations

import itertools
import os.path
import posixpath

from typing import Iterable

import dateutil.parser
from more_itertools import one

import jaraco.versioning as versioning
from jaraco.classes.ancestry import iter_subclasses


class Repo(versioning.VersionManagement):
    """
    An abstract class defining some interfaces for working with
    repositories.
    """

    def __init__(self, location='.'):
        self.location = location
        self.setup()

    def is_valid(self):
        "Return True if this instance is a valid for this location."
        return True

    def setup(self):
        pass

    @classmethod
    def get_valid_managers(cls, location):
        """
        Get the valid Repo implementations for this location.
        """

        def by_priority_attr(c):
            return getattr(c, 'priority', 0)

        classes = sorted(iter_subclasses(cls), key=by_priority_attr, reverse=True)
        instances = (c(location) for c in classes)
        return (inst for inst in instances if inst.is_valid())

    @staticmethod
    def detect(location='.'):
        try:
            return next(Repo.get_valid_managers(location))
        except StopIteration as e:
            e.args = ("No source repo or suitable VCS version found",)
            raise

    @staticmethod
    def existing_only(instances: Iterable[Repo]):
        """
        Return only those instances that refer to an existing repo
        """
        return (inst for inst in instances if inst.find_root())

    def __repr__(self):
        return '{self.__class__.__name__}({self.location})'.format(**vars())

    def find_root(self):
        raise NotImplementedError()

    def find_files(self):
        raise NotImplementedError()

    def get_tags(self, rev=None):
        """
        Get the tags for the specified revision (or the current revision
        if none is specified).
        """
        raise NotImplementedError()

    def get_repo_tags(self):
        """
        Get all tags for the repository.
        """
        raise NotImplementedError()

    def get_parent_tags(self, rev=None):
        """
        Return the tags for the parent revision (or None if no single
            parent can be identified).
        """
        try:
            parent_rev = one(self.get_parent_revs(rev))
        except Exception:
            return None
        return self.get_tags(parent_rev)

    def get_parent_revs(self, rev=None):
        """
        Get the parent revision for the specified revision (or the current
        revision if none is specified).
        """
        raise NotImplementedError

    def is_modified(self):
        'Does the current working copy have modifications'
        raise NotImplementedError()

    def find_all_files(self):
        """
        Find files including those in subrepositories.
        """
        files = self.find_files()
        subrepo_files = (
            posixpath.join(subrepo.location, filename)
            for subrepo in self.subrepos()
            for filename in subrepo.find_files()
        )
        return itertools.chain(files, subrepo_files)

    def subrepos(self):
        paths = (os.path.join(self.location, path) for path in self.sub_paths())
        return map(self.__class__, paths)

    def sub_paths(self):
        raise NotImplementedError()

    def get_timestamp(self, rev):
        return dateutil.parser.parse(self._get_timestamp_str(rev))

    def age(self):
        """
        Return the age of the repo.
        """
        raise NotImplementedError()

    def commit_tree(self, spec, msg="committed"):
        """
        Apply the tree in spec and commit.
        """
        raise NotImplementedError()



"""
>>> repo().get_current_version()
'...'
>>> isinstance(repo(), Repo)
True
"""

from .base import Repo
from .subprocess import Git, Mercurial

__all__ = ['Repo', 'repo', 'Mercurial', 'Git']

repo = Repo.detect



import os
import subprocess

from . import base, cmd


class Subprocess:
    env = None

    def _invoke(self, *params):
        """
        Invoke self.exe as a subprocess
        """
        cmd = [self.exe] + list(params)
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=self.location,
            env=self.env,
        )
        stdout, stderr = proc.communicate()
        if not proc.returncode == 0:
            raise RuntimeError(stderr.strip() or stdout.strip())
        return stdout.decode('utf-8')


class Mercurial(Subprocess, cmd.Mercurial, base.Repo):
    """
    A Repo implemented by calling into the 'hg' command-line
    as a subprocess.
    """

    priority = 1 + os.path.isdir('.hg')


class Git(Subprocess, cmd.Git, base.Repo):
    """
    A Repo implemented by calling into the 'git' command-line
    as a subprocess.
    """

    priority = 1 + os.path.isdir('.git')



import pytest

from .. import vcs


def _ensure_present(repo):
    try:
        repo.version()
    except Exception:
        pytest.skip()


@pytest.fixture
def temp_work_dir(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    return tmp_path


@pytest.fixture
def hg_repo(temp_work_dir):
    repo = vcs.Mercurial()
    _ensure_present(repo)
    repo._invoke('init', '.')
    return repo


@pytest.fixture
def git_repo(temp_work_dir):
    repo = vcs.Git()
    _ensure_present(repo)
    repo._invoke('init')
    repo._invoke('config', 'user.email', 'vip@example.com')
    repo._invoke('config', 'user.name', 'Important User')
    return repo



import abc
import collections
import itertools
import operator
import os.path
import re
import subprocess

import dateutil.parser
import jaraco.path
from tempora import utc


TaggedRevision = collections.namedtuple('TaggedRevision', 'tag revision')


class Command(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _invoke(self, *args): ...

    def is_valid(self):
        try:
            # Check if both command and repo are valid
            self._invoke('status')
        except Exception:
            return False
        return super().is_valid()

    def version(self):
        """
        Return the underlying version
        """
        lines = iter(self._invoke('version').splitlines())
        version = next(lines).strip()
        return self._parse_version(version)

    @classmethod
    def _parse_version(cls, version):
        return re.search(cls.version_pattern, version).group(1)


class Mercurial(Command):
    exe = 'hg'
    version_pattern = r'Mercurial Distributed SCM \((.*?)\)'

    def find_root(self):
        try:
            return self._invoke('root').strip()
        except Exception:
            pass

    def find_files(self):
        """
        Find versioned files in self.location
        """
        cmd = 'locate', '-I', '.', '--config', 'ui.relative-paths=yes'
        return self._invoke(*cmd).splitlines()

    def get_parent_revs(self, rev=None):
        cmd = ['parents', '--style', 'default', '--config', 'defaults.parents=']
        if rev:
            cmd.extend(['--rev', str(rev)])
        out = self._invoke(*cmd)
        cs_pat = r'^changeset:\s+(?P<local>\d+):(?P<hash>[0-9a-zA-Z]+)'
        return (match.groupdict()['local'] for match in re.finditer(cs_pat, out))

    def get_tags(self, rev=None):
        """
        Get the tags for the given revision specifier (or the
        current revision if not specified).
        """
        rev_num = self._get_rev_num(rev)
        # rev_num might end with '+', indicating local modifications.
        return (
            set(self._read_tags_for_rev(rev_num))
            if not rev_num.endswith('+')
            else set()
        )

    def _read_tags_for_rev(self, rev_num):
        """
        Return the tags for revision sorted by when the tags were
        created (latest first)
        """
        return (tr.tag for tr in self._read_tags_for_revset(rev_num))

    def _read_tags_for_revset(self, spec):
        """
        Return TaggedRevision for each tag/rev combination in the revset spec
        """
        cmd = ['log', '--style', 'default', '--config', 'defaults.log=', '-r', spec]
        res = self._invoke(*cmd)
        header_pattern = re.compile(r'(?P<header>\w+?):\s+(?P<value>.*)')
        match_res = map(header_pattern.match, res.splitlines())
        matched_lines = filter(None, match_res)
        matches = (match.groupdict() for match in matched_lines)
        for match in matches:
            if match['header'] == 'changeset':
                id, sep, rev = match['value'].partition(':')
            if match['header'] == 'tag':
                tag = match['value']
                yield TaggedRevision(tag, rev)

    def _get_rev_num(self, rev=None):
        """
        Determine the revision number for a given revision specifier.
        """
        # first, determine the numeric ID
        cmd = ['identify', '--num']
        # workaround for #4
        cmd.extend(['--config', 'defaults.identify='])
        if rev:
            cmd.extend(['--rev', rev])
        res = self._invoke(*cmd)
        return res.strip()

    def _get_tags_by_num(self):
        """
        Return a dictionary mapping revision number to tags for that number.
        """
        by_revision = operator.attrgetter('revision')
        tags = sorted(self.get_tags(), key=by_revision)
        revision_tags = itertools.groupby(tags, key=by_revision)

        def get_id(rev):
            return rev.split(':', 1)[0]

        return {
            get_id(rev): [tr.tag for tr in tr_list] for rev, tr_list in revision_tags
        }

    def get_repo_tags(self):
        lines = self._invoke('tags').splitlines()
        return (TaggedRevision(*line.rsplit(None, 1)) for line in lines if line)

    def get_ancestral_tags(self, rev='.'):
        """
        Like get_repo_tags, but only get those tags ancestral to the current
        changeset.
        """
        spec = 'sort(ancestors({rev}), -date)'.format(**vars())
        return self._read_tags_for_revset(spec)

    def is_modified(self):
        out = self._invoke('status', '-mard')
        return bool(out)

    def sub_paths(self):
        try:
            with open(os.path.join(self.location, '.hgsub')) as file:
                return [line.partition('=')[0].strip() for line in file]
        except Exception:
            return ()

    def _get_timestamp_str(self, rev):
        return self._invoke('log', '-l', '1', '--template', '{date|isodate}', '-r', rev)

    def commit_tree(self, spec, message: str = 'committed'):
        jaraco.path.build(spec)
        self._invoke('addremove')
        self._invoke('commit', '-m', message)


class Git(Command):
    exe = 'git'
    version_pattern = r'git version (\d+\.\d+[^ ]*)'

    def is_valid(self):
        return super().is_valid()

    def find_root(self):
        try:
            return self._invoke('rev-parse', '--top-level').strip()
        except Exception:
            pass

    def find_files(self):
        all_files = self._invoke('ls-files').splitlines()
        return all_files

    def get_tags(self, rev=None):
        """
        Return the tags for the current revision as a set
        """
        rev = rev or 'HEAD'
        return set(self._invoke('tag', '--points-at', rev).splitlines())

    def get_repo_tags(self):
        cmd = [
            "for-each-ref",
            "--sort=-committerdate",
            "--format=%(refname:short) %(objectname:short)",
            "refs/tags",
        ]

        lines = self._invoke(*cmd).splitlines()
        return (TaggedRevision(*line.rsplit(None, 1)) for line in lines if line)

    def is_modified(self):
        """
        Is the current state modified? (currently stubbed assuming no)
        """
        return False

    def get_ancestral_tags(self, rev=None):
        """
        Like get_repo_tags, but only get those tags ancestral to the current
        changeset.
        """
        rev = rev or 'HEAD'
        matches = self._invoke('tags', '--merged', rev).splitlines()
        return (rev for rev in self.get_repo_tags() if rev.tag in matches)

    def sub_paths(self):
        lines = self._invoke('submodules', 'status').splitlines()
        return (line.split()[1] for line in lines)

    def _get_timestamp_str(self, rev):
        return self._invoke('log', '-1', '--format=%ai', rev)

    def age(self):
        """
        >>> repo = getfixture('git_repo')
        >>> repo.age()
        datetime.timedelta(...)
        """
        proc = subprocess.Popen(
            ['git', 'log', '--reverse', '--pretty=%ad', '--date', 'iso'],
            stdout=subprocess.PIPE,
            text=True,
            encoding='utf-8',
        )
        first_line = proc.stdout.readline().strip()
        proc.terminate()
        proc.wait()
        proc.stdout.close()
        return utc.now() - dateutil.parser.parse(first_line)

    def commit_tree(self, spec, message: str = 'committed'):
        jaraco.path.build(spec)
        self._invoke('add', '.')
        self._invoke('commit', '-m', message)



import pytest


@pytest.fixture(autouse=True)
def _isolate_home(tmp_home_dir):
    """Isolate the tests from a developer's VCS config."""


rev1 = dict(
    bar=dict(
        baz="",
    ),
)


rev2 = dict(
    bar=dict(
        baz="content",
    ),
)


@pytest.fixture
def hg_repo(hg_repo):
    repo = hg_repo
    repo.commit_tree(rev1, 'committed')
    repo.commit_tree(rev2, 'added content')
    return repo


@pytest.fixture
def git_repo(git_repo):
    repo = git_repo
    repo.commit_tree(rev1, 'committed')
    repo.commit_tree(rev2, 'added content')
    return repo



from unittest import mock

import pytest

from jaraco import vcs


def test_existing_only():
    """
    Test the static method Repo.existing_only.

    Presumably, '/' is never an hg repo; at least for testing
    purposes, that's a reasonable assumption.
    """
    mgrs = vcs.Repo.get_valid_managers('/')
    existing = list(vcs.Repo.existing_only(mgrs))
    assert not existing


@mock.patch.object(
    vcs.Repo,
    'get_valid_managers',
    classmethod(lambda cls, location: iter(())),
)
def test_no_valid_implementations():
    """
    When no valid implementations can be found, a StopIteration is raised providing
    a nice message.
    """
    with pytest.raises(StopIteration) as err:
        vcs.Repo.detect()
    assert 'no source repo' in str(err).lower()



import sys

from . import base, cmd, reentry


class Mercurial(cmd.Mercurial, base.Repo):
    """
    A Repo implemented by invoking the hg command in-process.
    """

    def _invoke(self, *params):
        """
        Run the self.exe command in-process with the supplied params.
        """
        cmd = [self.exe, '-R', self.location] + list(params)
        with reentry.in_process_context(cmd) as result:
            sys.modules['mercurial.dispatch'].run()
        stdout = result.stdio.stdout.getvalue()
        stderr = result.stdio.stderr.getvalue()
        if not result.returncode == 0:
            raise RuntimeError(stderr.strip() or stdout.strip())
        return stdout.decode('utf-8')



import datetime
import operator
import os
import platform

import pytest

from jaraco import vcs
from jaraco.vcs import subprocess


@pytest.fixture(autouse=True)
def configure_username(monkeypatch):
    """
    Some environments (Windows on Azure) have Mercurial installed
    but not configured in a way that it can commit files, because
    the username isn't configured correctly. Monkeypatch the
    class to bypass this issue.
    """
    if platform.system() != 'Windows':
        return

    def _invoke(self, *params):
        params = ('--config', 'ui.username=tester') + params
        return super(subprocess.Mercurial, self)._invoke(*params)

    monkeypatch.setattr(subprocess.Mercurial, '_invoke', _invoke)


def test_subprocess_manager_invalid_when_exe_missing():
    """
    If the hg executable dosen't exist, the manager should report
    False for .is_valid().
    """
    non_existent_exe = '/non_existent_executable'
    assert not os.path.exists(non_existent_exe)
    mgr = subprocess.Mercurial()
    mgr.exe = non_existent_exe
    assert not mgr.is_valid()


@pytest.mark.usefixtures("hg_repo")
class TestRelativePaths:
    """
    Issue #9 demonstrated that the Manager would inadvertently return too many
    files when location is not the root of the repo. This test demonstrates
    the expected behavior.
    """

    def test_nested_child(self):
        test_mgr = vcs.Mercurial('.')
        assert test_mgr.find_files() == [os.path.join('bar', 'baz')]

    def test_manager_in_child(self):
        test_mgr = vcs.Mercurial('bar')
        assert test_mgr.find_files() == ['baz']

    def test_current_dir_in_child(self):
        os.chdir('bar')
        test_mgr = vcs.Mercurial('.')
        assert test_mgr.find_files() == ['baz']


@pytest.mark.usefixtures("hg_repo")
class TestTags:
    def setup_method(self, method):
        self.mgr = vcs.Mercurial('.')

    def teardown_method(self, method):
        del self.mgr

    def test_single_tag(self):
        self.mgr._invoke('tag', '1.0')
        assert self.mgr.get_tags() == {'tip'}
        self.mgr._invoke('update', '1.0')
        assert self.mgr.get_tags() == {'1.0'}

    def test_no_tags(self):
        "No tag should return empty set"
        assert self.mgr.get_tags('0') == set()

    def test_local_modifications(self):
        "Local modifications should return empty set"
        with open('bar/baz', 'w', encoding='utf-8') as f:
            f.write('changed')
        assert self.mgr.get_tags() == set()

    def test_parent_tag(self):
        self.mgr._invoke('tag', '1.0')
        assert self.mgr.get_tags() == {'tip'}
        assert self.mgr.get_parent_tags() == {'tip'}
        assert self.mgr.get_parent_tags('.') == {'1.0'}
        assert self.mgr.get_parent_tags('tip') == {'1.0'}
        self.mgr._invoke('tag', '1.1')
        assert self.mgr.get_tags() == {'tip'}
        assert self.mgr.get_parent_tags() == {'tip'}
        assert self.mgr.get_parent_tags('.') == {'1.1'}
        assert self.mgr.get_parent_tags('tip') == {'1.1'}

    def test_two_tags_same_revision(self):
        """
        Always return the latest tag for a given revision
        """
        self.mgr._invoke('tag', '1.0')
        self.mgr._invoke('tag', '-r', '1.0', '1.1')
        self.mgr._invoke('update', '1.0')
        assert set(self.mgr.get_tags()) == {'1.0', '1.1'}

    def test_two_tags_same_revision_lexicographically_earlier(self):
        """
        Always return the latest tag for a given revision
        """
        self.mgr._invoke('tag', '1.9')
        self.mgr._invoke('tag', '-r', '1.9', '1.10')
        self.mgr._invoke('update', '1.9')
        assert set(self.mgr.get_tags()) == {'1.9', '1.10'}

    def _setup_branchy_tags(self):
        """
        Create two heads, one which has a 1.0 tag and a different one which
        has a 1.1 tag.
        """
        # create a commit with a tag
        with open('bar/baz', 'a', encoding='utf-8') as f:
            f.write('\nmore to say\n')
        self.mgr._invoke('commit', '-m', 'Had more to say')
        self.mgr._invoke('tag', '1.0')
        # update to the pre-tagged revision
        self.mgr._invoke('update', '1')
        # Make a different commit
        with open('bar/baz', 'a', encoding='utf-8') as f:
            f.write('\na different concept\n')
        self.mgr._invoke('commit', '-m', 'A different approach')
        self.mgr._invoke('tag', '1.1')

    def test_ancestral_tags_local(self):
        """
        get_ancestral_tags should only return tagged revisions ancestral
        to the current revision.
        """
        self._setup_branchy_tags()
        tag_revs = self.mgr.get_ancestral_tags()
        tags = map(operator.attrgetter('tag'), tag_revs)
        # Two tags expected, one is the specified tag and
        # the other is 'tip'. On fast hardware, the
        # timestamps for these may be close enough that
        # mercurial doesn't sort them properly by time,
        # so normalize order.
        assert sorted(tags) == ['1.1', 'tip']

    def test_ancestral_tags_specified(self):
        """
        get_ancestral_tags should only return tagged revisions ancestral
        to the specified revision.
        """
        self._setup_branchy_tags()
        (tag,) = self.mgr.get_ancestral_tags(3)
        assert tag.tag == '1.0'


@pytest.mark.usefixtures("hg_repo")
class TestRevisionTimestamp:
    def test_tagged_rev_timestamp(self):
        mgr = vcs.Mercurial('.')
        mgr._invoke('tag', '1.0')
        assert mgr.get_timestamp('1.0').date() == datetime.date.today()



import sys

import pytest

from jaraco.vcs import reentry


def hello_world():
    print("hello world")


def hello_unicode_world():
    print(b"hello world".decode('ascii'))


def exit_zero():
    raise SystemExit(0)


def exit_one():
    raise SystemExit(1)


def exit_string():
    raise SystemExit('who does this?')


def echo():
    print("args are", sys.argv)


class TestReEntry:
    """
    Test the in-process entry point launcher
    """

    def test_hello_world(self):
        with reentry.in_process_context(['hello-world.py']) as proc:
            hello_world()
        assert proc.returncode == 0
        assert proc.stdio.stdout.getvalue() == 'hello world\n'

    def test_hello_world_unicode(self):
        with reentry.in_process_context(['hello-world.py']) as proc:
            hello_unicode_world()
        assert proc.returncode == 0
        assert proc.stdio.stdout.getvalue() == 'hello world\n'

    def test_main_with_system_exit(self):
        with reentry.in_process_context(['exit-zero']) as proc:
            exit_zero()
        assert proc.returncode == 0
        assert proc.stdio.stdout.getvalue() == ''

    def test_main_with_system_exit_one(self):
        with reentry.in_process_context(['exit-one']) as proc:
            exit_one()
        assert proc.returncode == 1
        assert proc.stdio.stdout.getvalue() == ''

    def test_main_with_system_exit_string(self):
        with reentry.in_process_context(['exit-string']) as proc:
            exit_string()
        assert proc.returncode == 1
        assert proc.stdio.stdout.getvalue() == ''

    def test_echo_args(self):
        with reentry.in_process_context(['echo', 'foo', 'bar']) as proc:
            echo()
        assert proc.returncode == 0
        out = "args are ['echo', 'foo', 'bar']\n"
        assert proc.stdio.stdout.getvalue() == out


class TestErrors:
    def test_name_error(self):
        with pytest.raises(NameError) as exc_info:
            with reentry.in_process_context([]) as proc:
                not_present  # noqa
        assert proc.returncode == 1
        msg = "name 'not_present' is not defined"
        assert str(exc_info.value).endswith(msg)

    def test_keyboard_interrupt(self):
        with pytest.raises(KeyboardInterrupt):
            with reentry.in_process_context([]) as proc:
                raise KeyboardInterrupt()
        assert proc.returncode == 1



import datetime
import os

import pytest

from jaraco import vcs
from jaraco.vcs import cmd, subprocess


def test_subprocess_manager_invalid_when_exe_missing():
    """
    If the executable dosen't exist, the manager should report
    False for .is_valid().
    """
    non_existent_exe = '/non_existent_executable'
    assert not os.path.exists(non_existent_exe)
    repo = subprocess.Git()
    repo.exe = non_existent_exe
    assert not repo.is_valid()


@pytest.mark.usefixtures("git_repo")
class TestTags:
    def test_single_tag(self):
        repo = vcs.Git('.')
        assert repo.get_tags() == set()
        repo._invoke('tag', '-am', "Tagging 1.0", '1.0')
        assert repo.get_tags() == {'1.0'}
        repo._invoke('checkout', '1.0')
        assert repo.get_tags() == {'1.0'}


class TestParseVersion:
    def test_simple(self):
        assert cmd.Git._parse_version('git version 1.9.3') == '1.9.3'

    def test_trailing_mess(self):
        val = cmd.Git._parse_version('git version 1.9.3 (Mac OS X)')
        assert val == '1.9.3'


@pytest.mark.usefixtures("git_repo")
class TestRevisionTimestamp:
    def test_tagged_rev_timestamp(self):
        repo = vcs.Git('.')
        repo._invoke('tag', '-am', 'tagging 1.0', '1.0')
        assert repo.get_timestamp('1.0').date() == datetime.date.today()


class TestIsolation:
    def test_commits_not_signed(self, git_repo):
        output = git_repo._invoke('log', '--show-signature')
        assert 'Signature made' not in output
